
import h5py as h5
import numpy as np

import logging as log

from ..core import DataModel
from ..lib.rag import create_rag
from ..lib.spencoding import spmeans
from ..lib.supersegments import HPC

DM = DataModel.instance()

def create_megavoxels(dataset=None, splabels=None, num_sps=None,
                      lamda=0.1, nbins=10, gamma=None, source=1,
                      out_mv=None, out_mvindex=None, out_mvtable=None):

    log.info('# Params: (num_sps: {}, lamda: {}, nbins: {}, gamma: {})'
             .format(num_sps, lamda, nbins, gamma))

    log.info('+ Loading data into memory')
    data = DM.load_slices(dataset)
    dshape = data.shape
    data.shape = (-1, 1)

    log.info('+ Loading supervoxels into memory')
    supervoxels = DM.load_slices(splabels)
    spsize = num_sps

    min_boundary = int(round(np.bincount(supervoxels.ravel()).mean() ** (1./3.)))
    log.info('+ Creating supervoxel graph: MinBound = {0:.2f}'.format(min_boundary))
    edges = create_rag(supervoxels, connectivity=6, min_boundary=10,
                       return_rag=False, return_counts=False)

    log.info('+ Creating supervoxel intensity features')
    data = spmeans(data, supervoxels)

    if gamma is not None:
        gamma = 1 / (2. * (gamma**2))

    log.info('+ Creating megavoxels')
    mvlabels = HPC(data, edges, lamda=lamda, nbow=nbins,
                   label_potential='l1', gamma=gamma,
                   min_size=None, max_size=spsize, verbose=False,
                   max_iter=2, nsamples=min(10000,spsize))

    total_mv = mvlabels.max()+1
    log.info('+ {} megavoxels created'.format(total_mv))

    mvlabels = mvlabels[supervoxels]

    log.info('+ Writing megavoxels to disk {}'.format(mvlabels.shape))
    params = {
        'lamda': lamda,
        'nbins': nbins,
        'gamma': 'None' if gamma is None else gamma,
        'num_megavoxels': total_mv,
        'source': dataset,
        'active': True
    }
    DM.create_empty_dataset(out_mv, shape=DM.data_shape, dtype=mvlabels.dtype)
    DM.write_slices(out_mv, mvlabels, params=params)

    log.info('+ Generating lookup table')
    mvravel = mvlabels.ravel()
    sortindexes = mvravel.argsort()
    sortlabels = mvravel[sortindexes]
    splits = np.searchsorted(sortlabels, np.arange(total_mv))

    if data.shape != DM.data_shape: # its a ROI, remap indexes
        log.info('+ Remaping indexes to ROI')
        zz, yy, xx = np.mgrid[:dshape[0], :dshape[1], :dshape[2]]
        zz += DM.active_roi[0].start
        yy += DM.active_roi[1].start
        xx += DM.active_roi[2].start
        seq = np.row_stack([zz.ravel(), yy.ravel(), xx.ravel()])
        idxtable = np.ravel_multi_index(seq, DM.data_shape)
        sortindexes = idxtable[sortindexes]

    log.info('+ Writing lookup table to disk {} {}'\
             .format(sortindexes.shape, splits.shape))
    DM.write_dataset(out_mvindex, sortindexes, params=dict(active=True))
    DM.write_dataset(out_mvtable, splits, params=dict(active=True))

    return out_mv, total_mv, out_mvindex, out_mvtable





from .data import load_data, volread
from .annotations import refine_label, save_threshold
from .supervoxels import create_supervoxels
from .megavoxels import create_megavoxels
from .channels import compute_channel, compute_all_channel
from .training import predict_proba
from .partition import label_objects, apply_rules
from .seganalysis import compare_segmentations



import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets

import logging as log

from collections import OrderedDict

from .base import Plugin
from ..widgets import HSize3D, HeaderLabel, SubHeaderLabel, \
                      HWidgets, RoundedWidget, ActionButton
from ..core import DataModel, Launcher, LayerManager
from .. import actions as ac


class ROIWidget(RoundedWidget):

    selected = QtCore.pyqtSignal(int)
    removed = QtCore.pyqtSignal(int)

    def __init__(self, idx, dfrom, dto, checked=False, removable=True, parent=None):
        super(ROIWidget, self).__init__(parent=parent, color=None, bg='#cde5e5',
                                        width=0)
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        self.idx = idx
        self.dfrom = dfrom
        self.dto = dto

        if removable:
            self.btn_rm = QtWidgets.QPushButton('X')
            self.btn_rm.clicked.connect(self.on_remove)
        else:
            self.btn_rm = QtWidgets.QWidget()
        self.btn_rm.setMinimumWidth(25)

        self.labelz = QtWidgets.QLabel('[{}, {})'.format(dfrom[0], dto[0]))
        self.labelz.setFixedWidth(67)
        self.labelz.setStyleSheet('color: #009999;')
        self.labely = QtWidgets.QLabel('[{}, {})'.format(dfrom[1], dto[1]))
        self.labely.setFixedWidth(67)
        self.labely.setStyleSheet('color: #009999;')
        self.labelx = QtWidgets.QLabel('[{}, {})'.format(dfrom[2], dto[2]))
        self.labelx.setFixedWidth(67)
        self.labelx.setStyleSheet('color: #009999;')

        self.z = QtWidgets.QLabel('z:')
        self.z.setStyleSheet('color: #6194BC; font-weight: bold;')
        self.z.setFixedWidth(7)
        self.y = QtWidgets.QLabel('y:')
        self.y.setStyleSheet('color: #6194BC; font-weight: bold;')
        self.y.setFixedWidth(7)
        self.x = QtWidgets.QLabel('x:')
        self.x.setStyleSheet('color: #6194BC; font-weight: bold;')
        self.x.setFixedWidth(7)

        self.chk = QtWidgets.QCheckBox()
        self.chk.setChecked(checked)
        self.chk.stateChanged.connect(self.mousePressEvent)

        vbox.addWidget(HWidgets(self.btn_rm, self.z, self.labelz,
                                self.y, self.labely, self.x, self.labelx,
                                None, self.chk, stretch=[0,0,0,0,0,0,0,1,0]))

    def mousePressEvent(self, ev):
        self.setChecked(~self.isChecked())
        self.selected.emit(self.idx)

    def setChecked(self, value):
        self.chk.blockSignals(True)
        self.chk.setChecked(value)
        self.chk.blockSignals(False)

    def isChecked(self):
        return self.chk.isChecked()

    def on_selected(self):
        self.selected.emit(self.idx)

    def value(self):
        slice_z = slice(self.dfrom[0], self.dto[0])
        slice_y = slice(self.dfrom[1], self.dto[1])
        slice_x = slice(self.dfrom[2], self.dto[2])
        return slice_z, slice_y, slice_x

    def on_remove(self):
        self.removed.emit(self.idx)


class ROI(Plugin):

    name = 'Select ROI'

    def __init__(self, parent=None):
        super(ROI, self).__init__(ptype=Plugin.Plugin, parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()

        self.addWidget(HeaderLabel('Select Region of Interest'))

        dummy = QtWidgets.QWidget(self)
        self.addWidget(dummy)

        vbox = QtWidgets.QVBoxLayout(self)
        dummy.setLayout(vbox)

        shape = self.DM.data_shape

        self.lbl_size = QtWidgets.QLabel('<b>Shape:</b> ' + str(shape))

        self.txt_from = HSize3D('From', default=(0, 0, 0), txtwidth=50)
        self.txt_to = HSize3D('To', default=shape, txtwidth=50)

        vbox.addWidget(self.lbl_size)
        vbox.addWidget(self.txt_from)
        vbox.addWidget(self.txt_to)

        self.btn_add = ActionButton('Add')
        vbox.addWidget(HWidgets(None, self.btn_add, stretch=[1,0]))

        vbox.addWidget(SubHeaderLabel('Available ROI'))

        self.container = QtWidgets.QWidget(self)
        self.container.setLayout(QtWidgets.QVBoxLayout(self))
        vbox.addWidget(self.container)

        self.rois = OrderedDict()
        self.nrois = 0
        self.addROI((0,0,0), shape, checked=True, removable=False)

        self.btn_add.clicked.connect(self.on_add_roi)

    def on_add_roi(self):
        dfrom = self.txt_from.value()
        dto = self.txt_to.value()
        self.addROI(dfrom, dto)
        self.saveROIs()

    def addROI(self, dfrom, dto, checked=False, removable=True, *args):
        idx = self.nrois
        roi = ROIWidget(idx, dfrom, dto, checked=checked, removable=removable)
        roi.selected.connect(self.on_select)
        roi.removed.connect(self.on_remove)

        self.rois[self.nrois] = roi
        self.container.layout().addWidget(roi)

        self.nrois += 1
        return idx

    def on_remove(self, idx):
        if self.rois[idx].isChecked():
            self.on_select(0)
        self.rois[idx].setParent(None)
        del self.rois[idx]
        self.saveROIs()

    def on_select(self, idx):
        for k, v in self.rois.items():
            if k != idx:
                v.setChecked(False)

        if not self.rois[idx].isChecked():
            self.rois[idx].setChecked(False)
            self.rois[0].setChecked(True)
            idx = 0

        self.DM.active_roi = self.rois[idx].value()
        self.DM.roi_changed.emit()
        self.LM.update()
        self.saveROIs()

    def saveROIs(self):
        rois = []
        for k, v in self.rois.items():
            if k == 0: continue
            roi = v.value()
            dfrom = [r.start for r in roi]
            dto = [r.stop for r in roi]
            rois.append(dfrom + dto + [int(v.isChecked())])
        self.DM.set_attrs('/data', dict(roi=rois))

    def loadROIs(self):
        rois = self.DM.attr('/data', 'roi')
        if rois is None:
            return
        selected = 0
        for roi in rois:
            dfrom = tuple(roi[:3])
            dto = tuple(roi[3:6])
            checked = roi[-1] == 1
            i = self.addROI(dfrom, dto, checked=checked)
            if checked:
                selected = i
        self.on_select(selected)





import numpy as np

import logging as log

from ..core import DataModel

from scipy.ndimage.measurements import label as splabel
from skimage.morphology import octahedron, ball

from scipy.ndimage import measurements as measure
from scipy.stats import binned_statistic
from sklearn.decomposition import PCA


def label_objects(dataset=None, labels=None, out=None, out_features=None,
                  source=None, return_labels=False):
    DM = DataModel.instance()

    log.info('+ Loading data into memory')
    data = DM.load_slices(dataset)
    if labels is None:
        data += 1
        labels = set(np.unique(data)) - set([0])
    else:
        data += 1
        labels = np.asarray(labels) + 1

    obj_labels = []

    log.info('+ Extracting individual objects')
    new_labels = np.zeros(data.shape, np.int32)
    total_labels = 0
    num = 0

    for label in labels:
        mask = (data == label)
        tmp_data = data.copy()
        tmp_data[~mask] = 0
        tmp_labels, num = splabel(tmp_data, structure=octahedron(1))
        mask = (tmp_labels > 0)
        new_labels[mask] = tmp_labels[mask] + total_labels
        total_labels += num
        obj_labels += [label] * num

    log.info('+ {} Objects found'.format(total_labels))
    log.info('+ Saving results')
    DM.create_empty_dataset(out, DM.data_shape, new_labels.dtype)
    DM.write_slices(out, new_labels, params=dict(active=True, num_objects=total_labels))

    log.info('+ Loading source to memory')
    data = DM.load_slices(source)
    objs = new_labels
    objects = new_labels
    num_objects = total_labels
    objlabels = np.arange(1, num_objects+1)

    log.info('+ Computing Average intensity')
    feature = measure.mean(data, objs, index=objlabels)
    DM.create_empty_dataset(out_features[0], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[0], feature, params=dict(active=True))

    """log.info('+ Computing Median intensity')
    objs.shape = -1
    data.shape = -1
    feature = binned_statistic(objs, data, statistic='median',
                               bins=num_objects+1)[0]
    feature = feature[objlabels]
    out_features[1].write_direct(feature)
    out_features[1].attrs['active'] = True
    objs.shape = dataset.shape
    data.shape = dataset.shape"""

    log.info('+ Computing Sum of intensity')
    feature = measure.sum(data, objs, index=objlabels)
    DM.create_empty_dataset(out_features[1], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[1], feature, params=dict(active=True))

    log.info('+ Computing Standard Deviation of intensity')
    feature = measure.standard_deviation(data, objs, index=objlabels)
    DM.create_empty_dataset(out_features[2], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[2], feature, params=dict(active=True))

    log.info('+ Computing Variance of intensity')
    feature = measure.variance(data, objs, index=objlabels)
    DM.create_empty_dataset(out_features[3], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[3], feature, params=dict(active=True))

    log.info('+ Computing Area')
    objs.shape = -1
    feature = np.bincount(objs, minlength=num_objects+1)[1:]
    DM.create_empty_dataset(out_features[4], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[4], feature, params=dict(active=True))
    DM.create_empty_dataset(out_features[5], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[5], np.log10(feature), params=dict(active=True))
    objs.shape = data.shape

    log.info('+ Computing Bounding Box')
    obj_windows = measure.find_objects(objs)
    feature = []; depth = []; height = []; width = [];
    for w in obj_windows:
        feature.append((w[0].stop - w[0].start) *
                       (w[1].stop - w[1].start) *
                       (w[2].stop - w[2].start))
        depth.append(w[0].stop - w[0].start)
        height.append(w[1].stop - w[1].start)
        width.append(w[2].stop - w[2].start)

    feature = np.asarray(feature, np.float32)
    DM.create_empty_dataset(out_features[6], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[6], feature, params=dict(active=True))
    #depth
    depth = np.asarray(depth, np.float32)
    DM.create_empty_dataset(out_features[7], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[7], depth, params=dict(active=True))
    # height
    height = np.asarray(height, np.float32)
    DM.create_empty_dataset(out_features[8], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[8], height, params=dict(active=True))
    # width
    width = np.asarray(width, np.float32)
    DM.create_empty_dataset(out_features[9], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[9], width, params=dict(active=True))
    # log10
    DM.create_empty_dataset(out_features[10], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[10], np.log10(feature), params=dict(active=True))

    log.info('+ Computing Oriented Bounding Box')
    ori_feature = []; ori_depth = []; ori_height = []; ori_width = [];
    for i, w in enumerate(obj_windows):
        z, y, x = np.where(objs[w] == i+1)
        coords = np.c_[z, y, x]
        if coords.shape[0] >= 3:
            coords = PCA(n_components=3).fit_transform(coords)
        cmin, cmax = coords.min(0), coords.max(0)
        zz, yy, xx = (cmax[0] - cmin[0] + 1,
                      cmax[1] - cmin[1] + 1,
                      cmax[2] - cmin[2] + 1)
        ori_feature.append(zz * yy * xx)
        ori_depth.append(zz)
        ori_height.append(yy)
        ori_width.append(xx)

    ori_feature = np.asarray(ori_feature, np.float32)
    DM.create_empty_dataset(out_features[11], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[11], ori_feature, params=dict(active=True))
    #depth
    ori_depth = np.asarray(ori_depth, np.float32)
    DM.create_empty_dataset(out_features[12], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[12], ori_depth, params=dict(active=True))
    # height
    ori_height = np.asarray(ori_height, np.float32)
    DM.create_empty_dataset(out_features[13], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[13], ori_height, params=dict(active=True))
    # width
    ori_width = np.asarray(ori_width, np.float32)
    DM.create_empty_dataset(out_features[14], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[14], ori_width, params=dict(active=True))
    # log10
    DM.create_empty_dataset(out_features[15], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[15], np.log10(ori_feature), params=dict(active=True))

    log.info('+ Computing Positions')
    pos = measure.center_of_mass(objs, labels=objs, index=objlabels)
    pos = np.asarray(pos, dtype=np.float32)
    DM.create_empty_dataset(out_features[16], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[16], pos[:, 2].copy(), params=dict(active=True))
    DM.create_empty_dataset(out_features[17], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[17], pos[:, 1].copy(), params=dict(active=True))
    DM.create_empty_dataset(out_features[18], (num_objects,), np.float32, check=False)
    DM.write_dataset(out_features[18], pos[:, 0].copy(), params=dict(active=True))

    if return_labels:
        return out, total_labels, np.asarray(obj_labels)

    return out, total_labels


def apply_rules(features=None, label=None, rules=None, out_ds=None, num_objects=None):
    DM = DataModel.instance()

    mask = np.ones(num_objects, dtype=np.bool)
    out = DM.load_ds(out_ds)
    out[out == label] = -1

    for f, s, t in rules:
        if s == 0:
            np.logical_and(mask, (DM.load_ds(features[f]) > t), out=mask)
        else:
            np.logical_and(mask, (DM.load_ds(features[f]) < t), out=mask)

    out[mask] = label
    DM.write_dataset(out_ds, out)
    return out




import numpy as np
from ..qt_compat import QtGui, QtCore
from collections import OrderedDict

from .model import DataModel
from .singleton import Singleton

from matplotlib import cm
import six

Axial = DataModel.instance().Axial
Sagittal = DataModel.instance().Sagittal
Coronal = DataModel.instance().Coronal


class Layer(object):

    def __init__(self, data, level=None, cmap=None, vmin=None, vmax=None, alpha=1., \
                 orient=Axial, background=None, threshold=None, pre=None,
                 binarize=False, visible=True):
        self.DM = DataModel.instance()
        self.data = data
        self.level = level
        self.orient = orient
        self.background = background
        self.cmap = cmap
        self.vmin = vmin
        self.vmax = vmax
        self.pre = pre
        self.alpha = alpha
        self.visible = visible
        self.index = -1
        self.threshold = threshold
        self.binarize = binarize

    def draw(self, ax, idx, i):
        self.index = i
        if self.visible:
            sslice = self.get_slice(self.data, idx)
            vmin, vmax = self.vmin, self.vmax
            if self.pre is not None:
                sslice = self.pre(sslice)
            if self.background is not None and self.cmap is not None:
                sslice = np.ma.masked_equal(sslice, self.background)
            elif self.threshold is not None and self.cmap is not None:
                target, thresh = self.threshold
                sslice = np.ma.masked_where(self.get_slice(target, idx) < thresh, sslice)
            elif self.binarize:
                mask = (sslice < vmin) | (sslice > vmax)
                sslice[mask] = 0
                sslice[~mask] = 1
                vmin, vmax = 0, 1
            ax.imshow(sslice, self.cmap, vmin=vmin, vmax=vmax,
                      alpha=self.alpha, interpolation='none')

    def update(self, image, idx):
        if self.visible:
            sslice = self.get_slice(self.data, idx)
            if self.pre is not None:
                sslice = self.pre(sslice)
            if self.background is not None and self.cmap is not None:
                sslice = np.ma.masked_equal(sslice, self.background)
            elif self.threshold is not None and self.cmap is not None:
                target, thresh = self.threshold
                sslice = np.ma.masked_where(self.get_slice(target, idx) < thresh, sslice)
            elif self.binarize:
                mask = (sslice < self.vmin) | (sslice > self.vmax)
                sslice[mask] = 0
                sslice[~mask] = 1
        else:
            mask = np.ones(self.shape(), np.bool)
            sslice = np.ma.masked_where(mask, mask)
        image.set_array(sslice)

    def shape(self):
        shape = list(self.data.shape)
        del shape[self.orient]
        return shape

    def get_slice(self, data, idx):
        if isinstance(data, six.string_types):
            return self.DM.load_slices(data, idx)[0]
        else:
            raise Exception('This shouldnt happen')


@Singleton
class LayerManager(QtCore.QObject):

    level_added = QtCore.pyqtSignal(str)
    added = QtCore.pyqtSignal(str, str)
    removed = QtCore.pyqtSignal(str, str)
    updated = QtCore.pyqtSignal()
    opacity = QtCore.pyqtSignal(str, str, float)
    toggled = QtCore.pyqtSignal(str, str, bool)

    def __init__(self):
        QtCore.QObject.__init__(self)
        self._levels = OrderedDict()

    def len(self):
        return len(self._layers)

    def levels(self):
        return list(self._levels.copy().keys())

    def layers(self, level=None):
        if level is None:
            return [val for lvl in self._levels.keys() for val in self._levels[lvl].values()]
        else:
            return [val for val in self._levels[level].values()]

    def layer_names(self, level=None):
        if level is None:
            return [val for lvl in self._levels.keys() for val in self._levels[lvl].keys()]
        else:
            return [val for val in self._levels[level].keys()]

    def visible_layers(self, level=None):
        if level is None:
            return [val for lvl in self._levels.keys() for val in self._levels[lvl].values() if val.visible]
        else:
            return [val for val in self._levels[level].values() if val.visible]

    def addLayer(self, data, name, level, **kwargs):
        if not level in self._levels:
            self._levels[level] = OrderedDict()
            self.level_added.emit(level)
        if not name in self._levels[level]:
            self._levels[level][name] = Layer(data, level, **kwargs)
            self.added.emit(name, level)
        else:
            self._levels[level][name] = Layer(data, level, **kwargs)

    def get(self, name, level):
        return self._levels[level][name]

    def remove(self, name, level):
        if name in self._levels[level]:
            del self._levels[level][name]
            self.removed.emit(name, level)

    def isin(self, name, level):
        return level in self._levels and str(name) in self._levels[level]

    def index(self, name, level):
        return self._levels[level][name].index if self._levels[level][name].visible else -1

    def clear(self, level=None):
        if level is None:
            self._levels.clear()
        else:
            self._levels[level].clear()

    def setOpacity(self, name, level, value):
        self._levels[level][name].alpha = value
        self.opacity.emit(name, level, value)

    def setVisible(self, name, level, bol):
        self._levels[level][name].visible = bol
        self.toggled.emit(name, level, bol)

    def setVMin(self, name, level, val):
        self._levels[level][name].vmin = val

    def setVMax(self, name, level, val):
        self._levels[level][name].vmax = val

    def update(self):
        self.updated.emit()
